/**
 * PWM library for avr
 * Supporting Timer0 (8bit), Timer1 (16bit) and Timer2 (8bit) 
 *
 * @author Amir Lazarovich
 * @version 0.1
 */

#ifndef __pwm_v1__
#define __pwm_v1__

#if defined(ARDUINO) && ARDUINO >= 100
        #include <Arduino.h>
#elif defined(ARDUINO)
        #include <WProgram.h>
#else
        #include "avr-io.h"
#endif

#include <stdlib.h>


#define ERROR_BAD_PIN -1
#define ERROR_NO_MAPPING -2
#define SUCCESS 1

#define TIMER_0 0
#define TIMER_1 1
#define TIMER_2 2

#define OC0A_PORT PORTD
#define OC0A_DIRECTION DDRD
#define OC0B_PORT PORTD
#define OC0B_DIRECTION DDRD
#define OC0A PIND6
#define OC0B PIND5

#define OC1A_PORT PORTB
#define OC1A_DIRECTION DDRB
#define OC1B_PORT PORTB
#define OC1B_DIRECTION DDRB
#define OC1A PINB1
#define OC1B PINB2

#define OC2A_PORT PORTB
#define OC2A_DIRECTION DDRB
#define OC2B_PORT PORTD
#define OC2B_DIRECTION DDRD
#define OC2A PINB3
#define OC2B PIND3

#define PWM_0 0
#define PWM_1 1
#define PWM_2 2
#define PWM_3 3
#define PWM_4 4
#define PWM_5 5 

#define TIMER_0_NO_PRESCALING (0 << CS02) | (0 << CS01) | (1 << CS00)
#define TIMER_0_DIVIDE_BY_8 (0 << CS02) | (1 << CS01) | (0 << CS00)
#define TIMER_0_DIVIDE_BY_64 (0 << CS02) | (1 << CS01) | (1 << CS00)
#define TIMER_0_DIVIDE_BY_256 (1 << CS02) | (0 << CS01) | (0 << CS00)
 #define TIMER_0_DIVIDE_BY_1024 (1 << CS02) | (0 << CS01) | (1 << CS00)

#define TIMER_1_NO_PRESCALING (0 << CS12) | (0 << CS11) | (1 << CS10)
#define TIMER_1_DIVIDE_BY_8 (0 << CS12) | (1 << CS11) | (0 << CS10)
#define TIMER_1_DIVIDE_BY_32 (0 << CS12) | (1 << CS11) | (1 << CS10)
#define TIMER_1_DIVIDE_BY_64 (1 << CS12) | (0 << CS11) | (0 << CS10)
#define TIMER_1_DIVIDE_BY_128 (1 << CS12) | (0 << CS11) | (1 << CS10)
#define TIMER_1_DIVIDE_BY_256 (1 << CS12) | (1 << CS11) | (0 << CS10)
#define TIMER_1_DIVIDE_BY_1024 (1 << CS12) | (1 << CS11) | (1 << CS10)

#define TIMER_2_NO_PRESCALING (0 << CS22) | (0 << CS21) | (1 << CS20)
#define TIMER_2_DIVIDE_BY_8 (0 << CS22) | (1 << CS21) | (0 << CS20)
#define TIMER_2_DIVIDE_BY_32 (0 << CS22) | (1 << CS21) | (1 << CS20)
#define TIMER_2_DIVIDE_BY_64 (1 << CS22) | (0 << CS21) | (0 << CS20)
#define TIMER_2_DIVIDE_BY_128 (1 << CS22) | (0 << CS21) | (1 << CS20)
#define TIMER_2_DIVIDE_BY_256 (1 << CS22) | (1 << CS21) | (0 << CS20)
#define TIMER_2_DIVIDE_BY_1024 (1 << CS22) | (1 << CS21) | (1 << CS20)

#define TIMER_0_MODE_PHASE_CORRECT (1 << WGM00)
#define TIMER_1_MODE_PHASE_AND_FREQUENCY_CORRECT (1 << WGM13)
#define TIMER_2_MODE_PHASE_CORRECT (1 << WGM20)

#define TCCRnA_MODE_MASK 0x03
#define TCCRnB_MODE_MASK 0x18

#define UNIT_A 0
#define UNIT_B 1
#define UNIT_A_AND_B 2

#define NUM_OF_PINS 6
int __pwmValueMapping[NUM_OF_PINS][4];


/* private */ int inline _pinToIndex(int pin);
void initTimer0(int outputCompareUnit, int prescaler, int mode);
void initTimer1(int outputCompareUnit, int prescaler, int mode);
void initTimer2(int outputCompareUnit, int prescaler, int mode);
void setTimer1Top(int value);
void initTimer(int timer, int outputCompareUnit, int prescaler, int mode);
void setPWMValueMapping(int pin, int minInput, int maxInput, int minOutput, int maxOutput);
int inline setPWMValue(int pin, int value);

// *********************** inline pwm.c ***********************
long map(long x, long in_min, long in_max, long out_min, long out_max) {
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

long constrain(long value, long min, long max) {
   if (value < min) {
      return min;
   } else if (value > max) {
      return max;
   } else {
      return value;
   }
}


int inline _pinToIndex(int pin) {
   switch(pin) {
      case PWM_0:
         return 0;

      case PWM_1:
         return 1;

      case PWM_2:
         return 2;

      case PWM_3:
         return 3;

      case PWM_4:
         return 4;

      case PWM_5:
         return 5;

      default:
         return -1;
   }
}

void initTimer0(int outputCompareUnit, int prescaler, int mode) {
   switch (outputCompareUnit) {
      case UNIT_A:
         TCCR0A = (1 << COM0A1) | (0 << COM0A0) | (mode & TCCRnA_MODE_MASK);
         setLow(OC0A_PORT, OC0A);
         setOutput(OC0A_DIRECTION, OC0A);
         break;

      case UNIT_B:
         TCCR0A = (1 << COM0B1) | (0 << COM0B0) | (mode & TCCRnA_MODE_MASK);
         setLow(OC0B_PORT, OC0B);
         setOutput(OC0B_DIRECTION, OC0B);
         break;

      case UNIT_A_AND_B:
         TCCR0A = (1 << COM0A1) | (0 << COM0A0) | (1 << COM0B1) | (0 << COM0B0) | (mode & TCCRnA_MODE_MASK);
         setLow(OC0A_PORT, OC0A);
         setLow(OC0B_PORT, OC0B);
         setOutput(OC0A_DIRECTION, OC0A);
         setOutput(OC0B_DIRECTION, OC0B);
         break;
   }

   TCCR0B = prescaler | (mode & TCCRnB_MODE_MASK);
}

void initTimer1(int outputCompareUnit, int prescaler, int mode) {
   switch (outputCompareUnit) {
      case UNIT_A:
         TCCR1A = (1 << COM1A1) | (0 << COM1A0) | (mode & TCCRnA_MODE_MASK);
         setLow(OC1A_PORT, OC1A);
         setOutput(OC1A_DIRECTION, OC1A);
         break;

      case UNIT_B:
         TCCR1A = (1 << COM1B1) | (0 << COM1B0) | (mode & TCCRnA_MODE_MASK);
         setLow(OC1B_PORT, OC1B);
         setOutput(OC1B_DIRECTION, OC1B);
         break;

      case UNIT_A_AND_B:
         TCCR1A = (1 << COM1A1) | (0 << COM1A0) | (1 << COM1B1) | (0 << COM1B0) | (mode & TCCRnA_MODE_MASK);
         setLow(OC1A_PORT, OC1A);
         setLow(OC1B_PORT, OC1B);
         setOutput(OC1A_DIRECTION, OC1A);
         setOutput(OC1B_DIRECTION, OC1B);
         break;
   }

   TCCR1B = prescaler | (mode & TCCRnB_MODE_MASK);
}

void initTimer2(int outputCompareUnit, int prescaler, int mode) {
   switch (outputCompareUnit) {
      case UNIT_A:
         TCCR2A = (1 << COM2A1) | (0 << COM2A0) | (mode & TCCRnA_MODE_MASK);
         setLow(OC2A_PORT, OC2A);
         setOutput(OC2A_DIRECTION, OC2A);
         break;

      case UNIT_B:
         TCCR2A = (1 << COM2B1) | (0 << COM2B0) | (mode & TCCRnA_MODE_MASK);
         setLow(OC2B_PORT, OC2B);
         setOutput(OC2B_DIRECTION, OC2B);
         break;

      case UNIT_A_AND_B:
         TCCR2A = (1 << COM2A1) | (0 << COM2A0) | (1 << COM2B1) | (0 << COM2B0) | (mode & TCCRnA_MODE_MASK);
         setLow(OC2A_PORT, OC2A);         
         setLow(OC2B_PORT, OC2B);
         setOutput(OC2A_DIRECTION, OC2A);
         setOutput(OC2B_DIRECTION, OC2B);
         break;
   }

   TCCR2B = prescaler | (mode & TCCRnB_MODE_MASK);
}

void setTimer1Top(int value) {
   ICR1 = value;
}

void initTimer(int timer, int outputCompareUnit, int prescaler, int mode) {
   switch (timer) {
      case TIMER_0:
         initTimer0(outputCompareUnit, prescaler, mode);
         break;

      case TIMER_1:
         initTimer1(outputCompareUnit, prescaler, mode);
         ICR1 = 25000; // 50 ms frequency -- how do we know the frequency? Fclk / (2 * N * TOP) -> N is the prescaler divider and TOP is the value we set on ICR1
         break;

      case TIMER_2:
         initTimer2(outputCompareUnit, prescaler, mode);
         break;
   }
}

void setPWMValueMapping(int pin, int minInput, int maxInput, int minOutput, int maxOutput) {
   int index = _pinToIndex(pin);
   if (index >= 0) {
      __pwmValueMapping[index][0] = minInput;
      __pwmValueMapping[index][1] = maxInput;
      __pwmValueMapping[index][2] = minOutput;
      __pwmValueMapping[index][3] = maxOutput;
   }
}

int inline setPWMValue(int pin, int value) {
   int index = _pinToIndex(pin);
   if (index < 0) {
      return ERROR_BAD_PIN;
   }

   // map value to specific range
   int minInput = __pwmValueMapping[index][0];
   int maxInput = __pwmValueMapping[index][1];
   int minOutput = __pwmValueMapping[index][2];
   int maxOutput = __pwmValueMapping[index][3];
   
   if (minOutput == 0 && maxOutput == 0) {
      return ERROR_NO_MAPPING;
      // throw "mapping not defined. Please call setPWMValueMapping before setting a value"
   }

   value = constrain(map(value, minInput, maxInput, minOutput, maxOutput), minOutput, maxOutput);
   switch (pin) {
      case PWM_0:
         OCR0A = value;
         break;

      case PWM_1:
         OCR0B = value;
         break;

      case PWM_2:
         OCR1A = value;
         break;

      case PWM_3:
         OCR1B = value;
         break;

      case PWM_4:
         OCR2A = value;
         break;

      case PWM_5:
         OCR2B = value;
         break;
   }

   return SUCCESS;
}
#endif