#!/usr/bin/python

from tornado.ioloop import IOLoop
import tornado.web as web
import tornadio2
import sys
import signal
import serial
import threading

DEBUG_ONLY_SERVER = True

APP_PORT = 8888
connections = []
isThreadActive = True


class DefaultWebHandler(web.RequestHandler):
    def get(self):
        self.render('controller.html')

class WebHandler(web.RequestHandler):
    def initialize(self):
        self.supported_path = ['controller.html', 'settings.html', 'controller-no-touch.html', 'favicon.ico']

    def prepare(self):
        action = self.request.path.split('/')[-1]
        action = self.fixUrl(action)
        if action not in self.supported_path:
            self.send_error(400)

    def get(self, page):
        self.render(self.fixUrl(page))

    def fixUrl(self, url):
        if ".html" not in url:
            url += ".html"
        return url


class EventHandler(tornadio2.SocketConnection):
	MOTOR_MIN = 0
	centerMotorPower = 0
	tailMotorPower = 0
	tailServoAngle = 0


	def on_open(self, request):
		print "client connected."
		connections.append(self)

	def on_close(self):
		print "connection closed."
		connections.remove(self)

	def emit_all(self, event, type, data):
		command = "[{event}] {type}:: data: {data}".format(event=event, type=type, data=data)
		print command

		serial.flush()
		if type == "standby":
			if data['on']:
				serial.write(" ") # stop all motors
		elif type == "sticks":
			val = max(data['throttle'], self.MOTOR_MIN)
			if val != self.centerMotorPower:
				self.centerMotorPower = val
				serial.write("m2 {}\n".format(self.centerMotorPower)) # center motor

			val = max(data['pitch'], self.MOTOR_MIN)
			if val != self.tailMotorPower:
				self.tailMotorPower = val
				serial.write("m1 {}\n".format(self.tailMotorPower)) # tail motor

			val = data['roll']
			if val != self.tailServoAngle:
				self.tailServoAngle = val
				serial.write("t1 {}\n".format(self.tailServoAngle)) # tail servo


	@tornadio2.event
	def control(self, **kwargs):
		self.emit_all("control", **kwargs)

	@tornadio2.event
	def function(self, **kwargs):
		self.emit_all("function", **kwargs)

	@tornadio2.event
	def log(self, **kwargs):
		self.emit_all("log", **kwargs)

	@tornadio2.event
	def response(self, **kwargs):
		self.emit_all("response", **kwargs)
	
	@tornadio2.event
	def settings(self, **kwargs):
		self.emit_all("settings", **kwargs)

	@tornadio2.event
	def get(self, **kwargs):
		self.emit_all("get", **kwargs)

class WebApp(object):
	def __init__(self):
		app_router = tornadio2.TornadioRouter(EventHandler)

		routes = [
			(r"/static/(.*)", web.StaticFileHandler, {"path": "./static"}),
            (r"/(?P<page>[^\/]+)", WebHandler),
            (r"/", DefaultWebHandler)
		]

		routes.extend(app_router.urls)

		self.application = web.Application(routes, socket_io_port = APP_PORT, debug = 1)

	def start(self, port = APP_PORT):
		self.application.listen(port)

# def serialLoop(serial):
# 	serial.flush()
    	# serial.read()

    	# while isThreadActive:
    		# TODO


def serialSignalHandler(signal, frame):
	isThreadActive = False
	print 'Serial signal detected! terminating program..'
	sys.exit(0)



if __name__ == "__main__":
	if not DEBUG_ONLY_SERVER:		
		if (len(sys.argv) != 3):
			print "\nusage: %s /path/to/serial baud-rate" % __file__
			print "example: %s /dev/tty.usbserial-FTGO091P 9600\n" % __file__
			sys.exit()

		serialPath = sys.argv[1]
		baudRate = sys.argv[2]

    		signal.signal(signal.SIGINT, serialSignalHandler)
    		signal.pause

 	   	serial = serial.Serial(serialPath, baudRate)
    		serial.setDTR()

    		# t = threading.Thread(target=serialLoop, args=(serial,))
   		# t.start()

    	WebApp().start()
    	IOLoop.instance().start()
